/* $Workfile: DspDemo2.c $   $Revision: 16 $    $Modtime: 2/03/03 4:06p $  */
/*--------------------------------------------------------------------------*\

  Odin TeleSystems Inc.

  SAA-1007-1 OTX DSP C54x Software Development Kit (SDK)

  Demo2

  This demo illustrates how to send and receive serial user data. It receives
  u-law encoded voice samples. It loops these samples back in the transmit
  direction after performing a gain adjustment. The gain adjustment value
  is selected by passing an IO Control code to this application.

  Use together with Host application Demo2.c

\*--------------------------------------------------------------------------*/

#include <string.h>   // for memset()

#include "Dsp54Api.h"
#include "Dsp54Ioctl.h"

#include "Demo2Io.h"

/*--------------------------------------------------------------------------*\
  Macros
\*--------------------------------------------------------------------------*/
#define DEMO_DATA_BUF_SIZE 25


/*--------------------------------------------------------------------------*\
  Private data
\*--------------------------------------------------------------------------*/

unsigned short g_hHandle = OTXDSP_INVALID_HANDLE;  // Handle passed to us from the host user application

static short g_law = OTXDSP_ULAW;

static short g_nTxGain = 0;

static short g_bRxEvenTimeSlot = OTXDSP_TRUE;   // Even or odd timeslot for the receiver
static short g_nRxWordSlot = 0;       // The "Word slot" (0-15) within a frame for the receiver

static short g_bTxEvenTimeSlot = OTXDSP_TRUE;   // Even or odd timeslot for the transmitted data
static short g_nTxWordSlot = 0;       // The "Word slot" (0-15) within a frame for the transmitted data

static short g_bRxEnabled = OTXDSP_FALSE;
static short g_bTxEnabled = OTXDSP_FALSE;

static short g_sqRxData[DEMO_DATA_BUF_SIZE];
static short g_sqTxData[DEMO_DATA_BUF_SIZE];
static short g_iTxData = 0;
static short g_iRxData = 0;

static short g_bStreamRxData = OTXDSP_FALSE;

Demo2NotifyGetDebugDataS g_debugDataS;

static volatile long  g_blinkCount      = 0;
static volatile long  g_blinkCountMax   = 100000L;
static volatile short g_blinkXfVal      = OTXDSP_FALSE;

/*--------------------------------------------------------------------------*\
  Function Prototypes
\*--------------------------------------------------------------------------*/
static void ProcessRxSamples(void);
static void GenerateTxSamples(void);
static void UnpackFrame(unsigned short *pFrame);
static void PackFrame(unsigned short *pFrame);
static void Blink(void);

/*--------------------------------------------------------------------------*\
  @func
    main() is called from _c_int00 which is the entry point for the boot loader.
    _c_int00 will first initialize all global data before running main().
\*--------------------------------------------------------------------------*/
main() {

  // Run Standard OTX DSP SDK API Initializations
  OtxDspInit();

  // Initialize application specific data
  memset(&g_debugDataS, 0, sizeof(Demo2NotifyGetDebugDataS));

  // Loop forever
  for(;;) {

    // check if we have RX samples available, process them. E.g. run detection algorithms on them
    if (g_bGotRxSerialPortInterrupt) {  // g_bGotRxSerialPortInterrupt updated by the interrupt service routine
      g_bGotRxSerialPortInterrupt = OTXDSP_FALSE;
      OtxDspServiceRxSerialPort();
    }

    // check if more samples are needed to be transmitted, generate more samples.
    if (g_bGotTxSerialPortInterrupt) {  // g_bGotTxSerialPortInterrupt updated by the interrupt service routine
      g_bGotTxSerialPortInterrupt = OTXDSP_FALSE;
      OtxDspServiceTxSerialPort();
    }

    // check if the host has requested our attention. If so, service his request
    if (g_bGotHostPortInterrupt) {  // g_bGotHostPortInterrupt updated by the interrupt service routine
      g_bGotHostPortInterrupt = OTXDSP_FALSE;
      OtxDspServiceIoControl();
    }

    // Blink the heartbeat LED to show that the DSP is up and running
    Blink();

  }  /* for */

}


/*--------------------------------------------------------------------------*\
  @func
    Blink the heart-beat LED
\*--------------------------------------------------------------------------*/
void Blink(
  void
)
{

  if (g_blinkCount++ >= g_blinkCountMax) {
    g_blinkCount = 0;

    if (g_blinkXfVal) {
      g_blinkXfVal = OTXDSP_FALSE;
      C54X_CLR_XF_BIT;   // Turn LED off
    } else {
      g_blinkXfVal = OTXDSP_TRUE;
      C54X_SET_XF_BIT;   // Turn LED on
    }
  }

}



/*--------------------------------------------------------------------------*\
    Call-back routine called by the Interrupt Service Routine when the
    host has issued a command to the HPI and requests our attention.

    This function is called from the idle loop in main()
    TO BE RE-IMPLEMENTED BY THE USER.
\*--------------------------------------------------------------------------*/
void OtxDspServiceIoControl(
  void
)
{
  //unsigned short nDataLen      = *((unsigned short *) OTXDSP_SDK_IOCTL_DATA_LEN_ADDR);
  unsigned short hHandle       = *((unsigned short *) OTXDSP_SDK_IOCTL_HANDLE_ADDR);
  unsigned short nTaskRef      = *((unsigned short *) OTXDSP_SDK_IOCTL_TASKREF_ADDR);
  //short          bDataComplete = *((unsigned short *) OTXDSP_SDK_IOCTL_CODE_ADDR) & OTXDSP_SDK_MASK_IO_DATA_COMPLETE;
  //short          bDataPadded   = *((unsigned short *) OTXDSP_SDK_IOCTL_CODE_ADDR) & OTXDSP_SDK_MASK_IO_DATA_PADDED;
  unsigned short nCode         = *((unsigned short *) OTXDSP_SDK_IOCTL_CODE_ADDR) & OTXDSP_SDK_MASK_IO_CODE;

  // Standard SDK IO Control codes
  if (nCode == OTXDSP_SDK_IO_CREATE) {
    if (g_hHandle == OTXDSP_INVALID_HANDLE) {
      // The handle is not set
      g_hHandle = hHandle;
      OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);
    } else {
      // This demo only supports one instance of the logical device. Return an error code.
      OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_E_OUT_OF_RESOURCES);
    }

  } else if (nCode == OTXDSP_SDK_IO_CLOSE) {
    g_bRxEnabled = OTXDSP_FALSE;
    g_bTxEnabled = OTXDSP_FALSE;
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_ENABLE) {
    g_bRxEnabled = OTXDSP_TRUE;
    g_bTxEnabled = OTXDSP_TRUE;
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_DISABLE) {
    g_bRxEnabled = OTXDSP_FALSE;
    g_bTxEnabled = OTXDSP_FALSE;
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_CONNECT) {
    OtxDspSdkIoInConnectS *pInS = (OtxDspSdkIoInConnectS *)OTXDSP_SDK_IOCTL_DATA_ADDR;
    short i;
    // Save these values in the debug struct. To be retrieved when the DSPDEMO1_IOCTL_GET_DEBUG_DATA control is issued
    g_debugDataS.nData0     = pInS->nDirection;

    if (pInS->nCapacity == 64) {
      // Only 64kbit/s connections are supported by this demo program
      for (i = 0; i < OTXDSP_SDK_WORDSLOTS_PER_HIGHWAY; i++) {
        if (pInS->sqConnectMask[i]) {
          break;
        }
      }

      if (pInS->nDirection == OTXDSP_SDK_RX_DIRECTION) {
        // Connection data for the Receiver
        g_nRxWordSlot = i;
        // Even or odd timeslot
        g_bRxEvenTimeSlot = (pInS->sqConnectMask[i] == 0x00FF) ? OTXDSP_TRUE : OTXDSP_FALSE;
      } else {
        // Connection data for the Transmitter
        g_nTxWordSlot = i;
        // Even or odd timeslot
        g_bTxEvenTimeSlot = (pInS->sqConnectMask[i] == 0x00FF) ? OTXDSP_TRUE : OTXDSP_FALSE;
      }

      g_debugDataS.nData1 = g_nRxWordSlot;
      g_debugDataS.nData2 = g_bRxEvenTimeSlot;
#if 1 // Temporary since we dont get ENABLE yet
      g_bRxEnabled = OTXDSP_TRUE;
      g_bTxEnabled = OTXDSP_TRUE;
#endif

      OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);
    } else {
      // Wrong Pipe capacity. Only 64kbit/s connections are supported by this demo program
      OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_E_INVALIDARG);
    }


  } else if (nCode == OTXDSP_SDK_IO_DISCONNECT) {
    // No specific action. Just disable.
    g_bRxEnabled = OTXDSP_FALSE;
    g_bTxEnabled = OTXDSP_FALSE;
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_RESET) {
    // No special reset action in this example. Just clear the debug data struct
    g_debugDataS.nData0 = 0;
    g_debugDataS.nData1 = 0;
    g_debugDataS.nData2 = 0;
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_INIT_BSP) {
    OtxDspInitBufferedSerialPorts(); 
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_START_PROGRAM) {
    unsigned short dspRc;
    dspRc = OtxDspStartProgram((OtxDspSdkIoInStartProgramS *)OTXDSP_SDK_IOCTL_DATA_ADDR);
    OtxDspCommandCompleteRc(hHandle, dspRc);
    
  } else if (nCode == OTXDSP_SDK_IO_INIT_TIMER) {
    unsigned short dspRc;
    dspRc = OtxDspInitTimer();
    OtxDspCommandCompleteRc(hHandle, dspRc);

    

  // Application specific IO Control Codes
  } else if (nCode == DEMO2_IOCTL_SET_TX_GAIN) {
    Demo2IoInSetTxGainS *pInS = (Demo2IoInSetTxGainS *)OTXDSP_SDK_IOCTL_DATA_ADDR;

    g_nTxGain = pInS->nTxGain;

    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == DEMO2_IOCTL_STREAM_RX_DATA) {
    Demo2IoInStreamRxDataS *pInS = (Demo2IoInStreamRxDataS *)OTXDSP_SDK_IOCTL_DATA_ADDR;

    g_bStreamRxData = pInS->bStreamRxData;

    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == DEMO2_IOCTL_GET_SAMPLE) {
    Demo2IoOutGetSampleS sOut;

    sOut.nResult = OTXDSP_SDK_S_OK;

    sOut.nDataSample = g_sqRxData[0];

    OtxDspCommandCompleteData(hHandle, (unsigned short *)&sOut, sizeof(Demo2IoOutGetSampleS));


  } else if (nCode == DEMO2_IOCTL_GET_DEBUG_DATA) {

    // This is an example of a non-blocking call
    // We must confirm the IoControl command before we can send any other events
    // back to the host.
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

    // A non-blocking call would normally take some time to execute.
    // Normally the application should save the Task Reference value, perform the task,
    // then call OtxDspSetHostEvent using the save Task Reference value.
    // However, in this example we call OtxDspSetHostEvent immediately.

    OtxDspSetHostEvent(DEMO2_NOTIFY_DEBUG_DATA, hHandle, nTaskRef, (unsigned short *)&g_debugDataS, sizeof(Demo2NotifyGetDebugDataS));

  } else {
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_E_INVALIDARG);
  }

}




void OtxDspServiceRxSerialPort(
/*--------------------------------------------------------------------------*\
    Call-back routine call when we have receive interrupt from the
    Buffered Serial Port indicating the the receive buffer is either full or
    half-full.

    This function is called from the idle loop in main()
    TO BE RE-IMPLEMENTED BY THE USER.
\*--------------------------------------------------------------------------*/
  void
)
{
  short n;
  unsigned short *pTxData;
  unsigned short *pRxData;


  if (*C54X_BSPCE_REG & 0x4000) {
    // BSPCE RH=1 --> We read the second half
    pRxData = (unsigned short *) (OTXDSP_BSP_RXBUF_ADDR + OTXDSP_BSP_RXBUF_SIZE/2);
    pTxData = (unsigned short *) (OTXDSP_BSP_TXBUF_ADDR + OTXDSP_BSP_TXBUF_SIZE/2);
  } else {
    // BSPCE RH=0 --> We read the first half
    pRxData = (unsigned short *) (OTXDSP_BSP_RXBUF_ADDR);
    pTxData = (unsigned short *) (OTXDSP_BSP_TXBUF_ADDR);
  }


  // At this point there are 25 frames of data to pick up
  // It is also time to fill the transmit buffer with data for 25 frames
  g_iTxData = 0;
  g_iRxData = 0;
  for (n = 0; n < 25; n++) {
    if (g_bRxEnabled) {
      UnpackFrame(pRxData + n*16);  // There are 16 words in one frame
    }
    if (g_bTxEnabled) {
      PackFrame(pTxData + n*16);
    }
  }

  ProcessRxSamples();
  GenerateTxSamples();

}


static void UnpackFrame(
/*--------------------------------------------------------------------------*\
  @func
    Picks up a certain timeslot from a frame, then expands the 8-bit timeslot
    to a 16-bit value using either u-law or a-law expansion (14-bit value with
    the 2 LSB set to 0).

    This function is called from OtxDspServiceRxSerialPort().

    Note that the TS0 octet is the most significant octet of the first
    unsigned short (16-bit) value in the frame. The TS1 octet is the least
    significant octet of the first unsigned short (16-bit) value in the frame.
    Therefore we must an 8 bit shift depending on if we want to pick up an even
    or an odd timeslot octet.

\*--------------------------------------------------------------------------*/
  unsigned short *pFrame
)
{
  //short n;
  unsigned short nTimeSlotOctet;

  g_debugDataS.nData0++;


  nTimeSlotOctet = *(pFrame + g_nRxWordSlot);
  if (g_bRxEvenTimeSlot) {
    // If we are working with an even timeslot (0,2,...,28,30) we must shift it down
    // to the Least Significant octet
    nTimeSlotOctet >>= 8;
  } else {
    // Odd timeslot. Just ignore the upper 8 bits to get the timeslot octet
    nTimeSlotOctet &= 0x00FF;
  }

  // Expand according to current law and store the data in our receive buffer
  g_sqRxData[g_iRxData++] = OtxDspLawExpand(nTimeSlotOctet, g_law);
}


static void PackFrame(
/*--------------------------------------------------------------------------*\
  @func

    Performs law compression on the outgoing data values. The compressed octet
    is then stored in the corresponding timeslot in the frame.

    This function is called from OtxDspServiceRxSerialPort().
    
\*--------------------------------------------------------------------------*/
  unsigned short *pFrame
)
{
  //short n;

  if (g_law == OTXDSP_ALAW) {
    // A-law
    if (g_bTxEvenTimeSlot) {
      // Even time slot. OR in the 8 bit compressed value in 8 MSB bits of the frame Word
      *(pFrame + g_nTxWordSlot) &= 0x00FF;
      *(pFrame + g_nTxWordSlot) |= ((OtxDspAlawCompress(g_sqTxData[g_iTxData++])) << 8);
    } else {
      // Odd time slot. OR in the 8 bit compressed value in 8 LSB bits of the frame Word
      *(pFrame + g_nTxWordSlot) &= 0xFF00;
      *(pFrame + g_nTxWordSlot) |= OtxDspAlawCompress(g_sqTxData[g_iTxData++]);
    }
  } else {
    // U-law
    if (g_bTxEvenTimeSlot) {
      // Even time slot. OR in the 8 bit compressed value in 8 MSB bits of the bch Word
      *(pFrame + g_nTxWordSlot) &= 0x00FF;
      *(pFrame + g_nTxWordSlot) |= ((OtxDspUlawCompress(g_sqTxData[g_iTxData++])) << 8);
    } else {
      // Odd time slot. OR in the 8 bit compressed value in 8 LSB bits of the bch Word
      *(pFrame + g_nTxWordSlot) &= 0xFF00;
      *(pFrame + g_nTxWordSlot) |= OtxDspUlawCompress(g_sqTxData[g_iTxData++]);
    }
  }

}



void OtxDspServiceTxSerialPort(
/*--------------------------------------------------------------------------*\
    Call-back routine call when we have transmit interrupt from the
    Buffered Serial Port indicating the the transmit buffer is either empty or
    half-empty.

    This function is called from the idle loop in main()
    TO BE RE-IMPLEMENTED BY THE USER.
\*--------------------------------------------------------------------------*/
  void
)
{
  // Performed by OtxDspServiceRxSerialPort() in this example
}


void ProcessRxSamples(
/*--------------------------------------------------------------------------*\
    Process Received Samples
\*--------------------------------------------------------------------------*/
  void
)
{
  if (g_bStreamRxData) {
    // Send the received data stream to the host
    static unsigned short nTaskRef = 0;
    Demo2NotifyStreamRxDataS streamRxDataS;

    memcpy(streamRxDataS.nData, g_sqRxData, 25);
    OtxDspSetHostEvent(DEMO2_NOTIFY_STREAM_RX_DATA, g_hHandle, nTaskRef++, (unsigned short *)&streamRxDataS, sizeof(Demo2NotifyStreamRxDataS));
  }
}


void GenerateTxSamples(
/*--------------------------------------------------------------------------*\
    Generate more samples to Transmit
\*--------------------------------------------------------------------------*/
  void
)
{
  short i;

  // Simply copy the received samples to the outgoing transmit data buffer
  // Perform a gain adjustment for each copied sample

  for (i = 0; i < DEMO_DATA_BUF_SIZE; i++) {
    g_sqTxData[i] = g_sqRxData[i] << g_nTxGain;
  }

}

